
import traceback
import time
import os.path
import linecache
import hilite_python

from style import stylesheet
from javascript import javascript

__all__ = ["TraceKnack", "TraceKnackError"]


#----------------------------------------------------------------------
# Configuration

SOURCE_RANGE = 4
TIME_OUT = 5 * 60

#----------------------------------------------------------------------
# Toolkit functions

def generate_ids():
    i = 0
    while True:
        i += 1
        yield i


def sanitise(s):
    return s.replace('&','&amp;').replace('<',"&lt;")


def sanitise_and_clip(s, l):
    if len(s)>l:
        return sanitise(s[:l-3])+"<span class='more'>...</span>"
    return sanitise(s)


def sanitise_pathnames(path):
    if ((path.startswith("/Library/Frameworks/Python") 
        or path.startswith("/usr/local/lib/")) 
        and "site-packages" in path):
        return path.split("site-packages",1)[1]
    if (path.startswith(os.getenv("HOME"))):
        return "~"+path[len(os.getenv("HOME")):]
    return path
        
    
def nice_type_name(instance):
    t = `type(instance)`
    if t.startswith("<type '") and t.endswith("'>"):
        return t[7:-2]
    return t


#----------------------------------------------------------------------
# Exceptions

class TraceKnackError(Exception):
    def __init__(self, original):
        self.internal_error = original
    def __str__(self):
        return "(TraceKnack Internal) "+str(self.internal_error)


def wrap_exceptions(func):
    """Decorator to cause exceptions inside the given function to be
    wrapped by TraceKnackError"""
    def wrapper(*args, **kw):
        try:
            return func(*args, **kw)
        except Exception, e:
            raise
            raise TraceKnackError(e)
    return wrapper
    

#----------------------------------------------------------------------
# Core functionality

def hilight_cache(fn, line, _cache={}):
    if line<0: return None
    try:
        html = _cache[fn]
    except KeyError:
        if fn=='<string>':
            html = "No source to view"
        else:
            try:
                html = "".join(list(hilite_python.generate_html(file(fn,"rU").read)))
            except IOError:
                html = "Unable to view source for '%s' due to I/O error" % sanitise(fn)
            except:
                html = "Unable to view source for '%s'" % sanitise(fn)
        _cache[fn] = html = [item+'\n' for item in html.split('\n')]
    if line>=len(html): return None
    return html[line]

def render_source(fn, line, source_range):
    line-=1
    lines = [(i,hilight_cache(fn, i)) for i in range(max(0,line-source_range), line+source_range+1)]
    result=[]
    for linenum, source in lines:
        if source is not None:
            source = "<span class='lineNum'>%d</span> %s" % (linenum, source)
            if linenum==line: source="<em>%s</em>" % source
            result.append(source)
    return "".join(result)
  
    
def get_sub_items(value, include_hidden=False, include_callables=False):
    if isinstance(value, list) or isinstance(value, tuple) or isinstance(value, dict):
        return value
    result = {}
    for key in dir(value):
        if not include_hidden and key.startswith('_'):
            continue
        v = getattr(value, key)
        if include_callables or not callable(v):
            result[key]=v
    return result


def get_items(container, callables, hidden):
    if isinstance(container, dict):
        return container.items()
    if isinstance(container, list) or isinstance(container, tuple):
        return enumerate(container)
    return get_sub_items(container, hidden, callables).items()
    
    
def format_variable((key, value)):
    if get_sub_items(value):
        klass = "folded hasChildren"
        extra = " onclick='toggleVisibility(this)'"
    else:
        klass = "noChildren"
        extra = ""
    return "<li class='%s'%s><var>%s</var> <code>%s</code> <samp>%s</samp></li>" % (
        klass, extra, key, 
        sanitise_and_clip(nice_type_name(value),200),
        sanitise_and_clip(`value`,200)
    ) 
 
 
def descend_into(container, key):
    if isinstance(container, dict):
        return container.get(key)
    elif isinstance(container, list) or isinstance(container, tuple):
        return container[int(key)]
    else:
        return getattr(container, key)
    

#----------------------------------------------------------------------
# Entry points    

class TraceKnack(object):
    def __init__(self, root_path, source_range=SOURCE_RANGE):
        self.root_path = root_path
        self.source_range = source_range
        self.excepts = {}
        self.ids = generate_ids()        

    def handle(self, (exc_class, exc, trace)):
        """Handle an exception.
        Pass in the results of sys.exc_info() during exception-handling. Returns a
        complete HTML document displaying the exception and allowing browsing of the
        stack, variables and a window of the source-code.
        """
        self.clean_old_exceptions()
        newID = self.ids.next()
        self.excepts[newID] = [exc, trace, time.time()]
        return self.render_viewer(newID)
        
    def _render_toolbar(self, depth, trace_list, include_hidden=False, include_callables=False):
        path_root = os.path.commonprefix([sanitise_pathnames(item[0]) for item in trace_list])
        stack_frames = ["<option>&lt;Globals&gt;</option>"] + \
            ["<option%s>%s line %d in %s()</option>" % \
                ([""," selected"][i==depth],  sanitise_pathnames(fn)[len(path_root):], line, fun) 
                    for (i,(fn, line, fun, src)) in enumerate(trace_list)
            ]
        stack_frames = "\r\n".join(stack_frames)
        checked=["","checked"]
        return """
        <form class="toolbar">
            %s <select id="stackSelector" onChange="setStackDepth()">
            %s
            </select>
            <input type="checkbox" id="callables" onChange="setStackDepth()" %s /> Include methods
            <input type="checkbox" id="hidden" onChange="setStackDepth()" %s /> Include hidden
        </form>
        """ % (path_root, stack_frames, checked[include_callables], checked[include_hidden])
        
    def _get_frame(self, exc, trace, trace_list, depth):
        if depth is None:
            return None, None, trace.tb_frame.f_globals
        else:
            if depth<0:
                depth+=len(trace_list)
            if (depth>=0) and (depth<len(trace_list)):
                this_frame = trace
                d = depth
                while d>0:
                    this_frame = this_frame.tb_next
                    d-=1
                if this_frame:
                    return depth, this_frame, this_frame.tb_frame.f_locals
            return depth, None, {}
        
    @wrap_exceptions
    def get_variables(self, exceptID, depth, callables, hidden, path):
        """Return the variables of a stack frame as a fragment of HTML.
        For a given exception ID and stack depth, find the given 'path' (eg x.y.z) and
        returns all members of the structure found there. 'callables' and 'hidden'
        are booleans allowing or hiding call()able items and _masked entries. 'depth'
        is treated like a list index -- starts from zero and goes upwards, but negative
        numbers are treated as working backwards from the end. A depth of 'None' ignores
        the stack and returns the global variables instead. Variables are returned as 
        html list items (<li> tags).
        """
        data = self.excepts.get(exceptID)
        if not data:
            return "null"
        data[2] = time.time()
        exc, trace, tstamp = data
        trace_list = traceback.extract_tb(trace)
        frameVars = self._get_frame(exc, trace, trace_list, depth)[2]
        path = path.split(".")
        for item in path:
            frameVars = descend_into(frameVars, item)
            if not frameVars:
                return "null"
        varList = "\r\n".join( map(format_variable, get_items(frameVars, callables, hidden)) )
        return varList
        
    @wrap_exceptions
    def render_viewer(self, exceptID, depth=-1, callables=False, hidden=False):
        """Render an HTML document for a given exception ID providing a debugging
        interface that displays the code, variables and call stack.
        'callables' and 'hidden' are booleans allowing or hiding call()able items and
        _masked entries. 'depth' is treated like a list index -- starts from zero and 
        goes upwards, but negative numbers are treated as working backwards from the end. 
        A depth of 'None' ignores the stack and returns the global variables instead. 
        Variables are returned as html list items (<li> tags).
        """
        data = self.excepts.get(exceptID)
        if not data:
            return "No "+str(exceptID)+" known here."+`self.excepts`
        data[2] = time.time()
        exc, trace, tstamp = data
        traceText = traceback.format_exception_only(exc.__class__, exc)
        traceTitle = traceText[0]
        trace_list = traceback.extract_tb(trace)
        depth, thisFrame, frameVars = self._get_frame(exc, trace, trace_list, depth)
        if depth is None:
            code = ""
        else:
            code = render_source(trace_list[depth][0], trace_list[depth][1], self.source_range)            
        if len(traceText)>1:
            traceTitle+=" <small>%s</small>" % " ".join(traceText[1:])
        title = "TraceKnack"
        header = """
        <div class="header">
            <h1>%s</h1>
            <h2>%s</h2>
            %s
        </div>
        """ % (traceTitle, time.strftime("%c"), self._render_toolbar(depth, trace_list, hidden, callables))
        varList = "\r\n".join( map(format_variable, frameVars.items()) )
        content = """
        <div class="output">
            <code>%s</code>
            <ul class="inspector">
                %s
            </ul>
        </div>
        """ % (code, varList)
        config = """
        VIEWER_URL = "%sViewer?"
        AJAX_URL = "%sVar?"
        EXCEPT_ID = %d
        """ % (self.root_path, self.root_path, exceptID)
        js = javascript % config

        return """<html>
        <head><title>%s</title>
        %s
        %s
        </head>
        <body>
        %s
        %s
        </body>""" % (title, stylesheet, js, header, content)
        
    def clean_old_exceptions(self):
        """Remove exceptions from the cache if they haven't been accessed for
        longer than the time-out period. Called automatically whenever a new
        exception occurs."""
        now = time.time()
        kill_list=[]
        for (key, (_, _, tstamp)) in self.excepts.items():
            if (now-tstamp)>TIME_OUT:
                kill_list.append(now)
        for key in kill_list:
            del self.excepts[key]

